/*
 * Local Matters Voice and Wireless Division, Inc. Confidential
 * OCO Source Materials
 * Copyright (C) Local Matters Voice and Wireless Division, Inc. 2007.  All Rights Reserved.
 *
 * The source code for this program is not published or otherwise divested of its trade
 * secrets, irrespective of what has been deposited with the U.S. Copyright Office.
 */
package com.localmatters.collector.schema;

import com.localmatters.collector.schema.entity.CollectorSchema;
import com.localmatters.collector.schema.entity.CollectorSchemaPK;
import java.util.ArrayList;
import java.util.List;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import javax.ejb.Stateless;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The schema-manager service implementation.
 *
 * Upon bean initialization, the default schema is loaded. The default schema
 * is identified by a default schema name and version. These values are injected
 * upon the bean creation and are defined in the ejb-jar.xml descriptor file as
 * default-schema-name and default-schema-version respectively. If either of
 * these values are missing or the schema does not exist, proper messages are
 * logged into configured logging facility.
 *
 * @author Daniel Cavalcanti
 */
@Stateless()
@WebService()
public class SchemaBean
    implements SchemaRemote, SchemaLocal {
    
    // **********************************************************************
    // Class constants
    // **********************************************************************
    
    private static final String SCHEMA_FIND_ALL = "Schema.findAll";
    private static final String SCHEMA_FIND_ALL_NAMES = "Schema.findAllNames";
    private static final String SCHEMA_FIND_BY_NAME = "Schema.findByName";
    private static final String SCHEMA_NAME_PARAM = "name";
    
    // **********************************************************************
    // Class variables
    // **********************************************************************
    
    /**
     * The class logger.
     */
    private static Log log = LogFactory.getLog(SchemaBean.class);
    
    // **********************************************************************
    // Instance variables
    // **********************************************************************
    
    /**
     * The persistent context entity manager.
     */
    @PersistenceContext(unitName="collector-persistence")
    private EntityManager manager;
    
    /**
     * The default schema name.
     */
    @Resource(name="default-schema-name")
    private String defaultSchemaName;
    
    /**
     * The default schema version.
     */
    @Resource(name="default-schema-version")
    private String defaultSchemaVersion;
    
    /**
     * The default schema.
     */
    private String defaultSchema;
    
    // **********************************************************************
    // Constructors
    // **********************************************************************
    
    /**
     * Creates a new instance of SchemaBean.
     */
    public SchemaBean() {
        log.debug("Creating schema bean.");
    }
    
    // **********************************************************************
    // Life-cycle methods
    // **********************************************************************
    
    /**
     * Initializes service.
     */
    @PostConstruct()
    private void init() {
        
        log.info("Initializing schema bean.");
        
        // ******************************************************************
        // Load default schema
        // ******************************************************************
        
        try {
            
            log.debug("Loading default schema.");
            
            if (defaultSchemaName == null)
                log.error("Default schema name is missing. Cannot load default schema.");
            if (defaultSchemaVersion == null)
                log.error("Default schema version is missing. Cannot load default schema.");
            
            if (defaultSchemaName != null && defaultSchemaVersion != null) {
                
                CollectorSchemaPK pk = new CollectorSchemaPK(defaultSchemaVersion, defaultSchemaName);
                CollectorSchema entity = manager.find(CollectorSchema.class, pk);
                
                if (entity != null) {
                    log.debug("Default schema found in database.");
                    log.trace(entity.getDocument());
                    defaultSchema = entity.getDocument();
                } else {
                    log.error("Default schema not found in database.");
                }
                
            }
            
        } catch (Throwable ex) {
            log.error("Unexpected error. Bean may not have been initialized properly.", ex);
        }
        
    }
    
    /**
     * Release allocated resources.
     */
    @PreDestroy()
    private void release() {
        log.info("Releasing resources in schema bean.");
    }
    
    // **********************************************************************
    // SchemaService methods
    // **********************************************************************
    
    @WebMethod()
    public void addSchema(@WebParam() String name, @WebParam() String version, @WebParam() String schema)
    throws NullPointerException, InvalidSchemaException, SchemaException {
        
        try {
            
            log.info("Requesting to add schema to the database.");
            
            // **************************************************************
            // validate parameters
            // **************************************************************
            
            if (name == null)
                throw new NullPointerException("Schema name cannot be null.");
            if (version == null)
                throw new NullPointerException("Schema version cannot be null.");
            if (schema == null)
                throw new NullPointerException("Schema cannot be null.");
            
            if (!valid(schema))
                throw new InvalidSchemaException("Schema not valid.", new IllegalArgumentException(schema));
            
            // **************************************************************
            // log parameters
            // **************************************************************
            
            final String endl = System.getProperty("line.separator", "\n");
            StringBuilder buf = new StringBuilder(endl);
            
            buf.append("schema name....: ").append(name).append(endl);
            buf.append("schema version.: ").append(version).append(endl);
            buf.append(schema);
            
            log.trace(buf.toString());
            
            // **************************************************************
            // persist schema
            // **************************************************************
            
            persist(name, version, schema);
            
        } catch (InvalidSchemaException ex) {
            throw ex;
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    @WebMethod()
    public List<String> getNames()
    throws SchemaException {
        
        try {
            
            log.info("Requesting schema names.");
            
            // **************************************************************
            // search database
            // **************************************************************
            
            log.trace("Searching database.");
            
            Query query = manager.createNamedQuery(SCHEMA_FIND_ALL_NAMES);
            List<String> names = query.getResultList();
            
            log.trace("Found " + names.size() + " schema names.");
            
            return names;
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    @WebMethod()
    public List<String> getVersions(@WebParam() String name)
    throws NullPointerException, SchemaException {
        
        try {
            
            log.info("Requesting versions for schema " + name);
            
            // **************************************************************
            // validate parameters
            // **************************************************************
            
            if (name == null)
                throw new NullPointerException("Schema name cannot be null.");
            
            // **************************************************************
            // search database
            // **************************************************************
            
            log.trace("Searching database.");
            
            Query query = manager.createNamedQuery(SCHEMA_FIND_BY_NAME);
            query.setParameter(SCHEMA_NAME_PARAM, name);
            
            List<CollectorSchema> entities = query.getResultList();
            
            log.trace("Found " + entities.size() + " versions for schema " + name);
            
            // **************************************************************
            // build version set
            // **************************************************************
            
            log.trace("Building versions set.");
            
            List<String> versions = new ArrayList<String>(entities.size());
            for (CollectorSchema entity : entities)
                versions.add(entity.getSchemaPK().getVersion());
            
            return versions;
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    @WebMethod()
    public String getSchema(@WebParam() String name, @WebParam() String version)
    throws NullPointerException, SchemaException {
        
        try {
            
            log.info("Requesting schema " + name + " <" + version + ">");
            
            // **************************************************************
            // validate parameters
            // **************************************************************
            
            if (name == null)
                throw new NullPointerException("Schema name cannot be null.");
            if (version == null)
                throw new NullPointerException("Schema version cannot be null.");
            
            // **************************************************************
            // search database
            // **************************************************************
            
            log.trace("Searching database.");
            
            CollectorSchemaPK pk = new CollectorSchemaPK(version, name);
            CollectorSchema entity = manager.find(CollectorSchema.class, pk);
            
            if (entity != null)
                log.trace("Found schema in database.\n" + entity.getDocument());
            else
                log.trace("Schema not found in database.");
            
            return (entity != null ? entity.getDocument() : null);
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    @WebMethod()
    public List<String> getAllSchemas()
    throws SchemaException {
        
        try {
            
            log.info("Requesting all schemas.");
            
            // **************************************************************
            // search database
            // **************************************************************
            
            log.trace("Searching database.");
            
            Query query = manager.createNamedQuery(SCHEMA_FIND_ALL);
            List<CollectorSchema> entities = query.getResultList();
            
            log.trace("Found " + entities.size() + " schemas.");
            
            // **************************************************************
            // build schema set
            // **************************************************************
            
            log.trace("Building schemas set.");
            
            List<String> schemas = new ArrayList<String>(entities.size());
            for (CollectorSchema entity : entities)
                schemas.add(entity.getDocument());
            
            return schemas;
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    @WebMethod()
    public void updateSchema(@WebParam() String name, @WebParam() String version, @WebParam() String schema)
    throws NullPointerException, InvalidSchemaException, SchemaException {
        
        try {
            
            log.info("Requesting to update schema " + name + " <" + version + ">");
            
            // **************************************************************
            // validate parameters
            // **************************************************************
            
            if (name == null)
                throw new NullPointerException("Schema name cannot be null.");
            if (version == null)
                throw new NullPointerException("Schema version cannot be null.");
            if (schema == null)
                throw new NullPointerException("Schema cannot be null.");
            
            if (!valid(schema))
                throw new InvalidSchemaException("Schema not valid.", new IllegalArgumentException(schema));
            
            // **************************************************************
            // log parameters
            // **************************************************************
            
            final String endl = System.getProperty("line.separator", "\n");
            StringBuilder buf = new StringBuilder(endl);
            
            buf.append("schema name....: ").append(name).append(endl);
            buf.append("schema version.: ").append(version).append(endl);
            buf.append(schema);
            
            log.trace(buf.toString());
            
            // **************************************************************
            // persist schema
            // **************************************************************
            
            persist(name, version, schema);
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    @WebMethod()
    public void removeSchema(@WebParam() String name, @WebParam() String version)
    throws NullPointerException, SchemaException {
        
        try {
            
            log.info("Requesting to remove schema " + name + " <" + version + ">");
            
            // **************************************************************
            // validate parameters
            // **************************************************************
            
            if (name == null)
                throw new NullPointerException("Schema name cannot be null.");
            if (version == null)
                throw new NullPointerException("Schema version cannot be null.");
            
            // **************************************************************
            // search database
            // **************************************************************
            
            log.trace("Searching database.");
            
            CollectorSchemaPK pk = new CollectorSchemaPK(version, name);
            CollectorSchema entity = manager.find(CollectorSchema.class, pk);
            
            if (entity != null) {
                log.trace("Found schema in database.\n" + entity.getDocument());
                manager.remove(entity);
            } else
                log.trace("Schema not found in database.");
            
        } catch (Throwable ex) {
            log.error("Unexpected error.", ex);
            throw new SchemaException(ex);
        }
        
    }
    
    // **********************************************************************
    // Auxiliary methods
    // **********************************************************************
    
    /**
     * Persists a schema in the database.
     *
     * @param name The schema name.
     * @param version The schema version.
     * @param schema The schema.
     */
    private void persist(String name, String version, String schema) {
        
        log.trace("Persisting schema.");
        
        CollectorSchemaPK pk = new CollectorSchemaPK(version, name);
        CollectorSchema entity = manager.find(CollectorSchema.class, pk);
        
        if (entity == null) {
            
            log.debug("New schema. adding to database.");
            entity = new CollectorSchema(pk, schema);
            manager.persist(entity);
            
        } else {
            
            log.debug("Schema already exists. Updating database.");
            entity.setDocument(schema);
            manager.merge(entity);
            
        }
        
    }
    
    /**
     * Validates a schema.
     *
     * @return true iff schema is valid; otherwise, false.
     * @param schema The schema to validate.
     */
    private boolean valid(String schema) {
        // TODO validate schema
        return true;
    }
    
}
